/*
 * \brief   fulda - use ieee1394 for debugging.
 * \date    2007-04-19
 * \author  Bernhard Kauer <kauer@tudos.org>
 */
/*
 * Copyright (C) 2007  Bernhard Kauer <kauer@tudos.org>
 * Technische Universitaet Dresden, Operating Systems Research Group
 *
 * This file is part of the FULDA package, which is distributed under
 * the  terms  of the  GNU General Public Licence 2.  Please see the
 * COPYING file for details.
 */

#include "dev.h"
#include "mbi.h"
#include "util.h"
#include "version.h"

const char *message_label = "Fulda: ";

enum
  {
    ARTRY     =   0x8 / sizeof(unsigned),
    BOPT      =  0x20 / sizeof(unsigned),
    HCC_set   =  0x50 / sizeof(unsigned),
    HCC_clear =  0x54 / sizeof(unsigned),
    LC_clear  =  0xe4 / sizeof(unsigned),
    ASRFH_set = 0x100 / sizeof(unsigned),
    PHRFH_set = 0x110 / sizeof(unsigned),
    PHRFL_set = 0x118 / sizeof(unsigned),
    PHUP      = 0x120 / sizeof(unsigned),
    ATRQ_set  = 0x180 / sizeof(unsigned),
    ATRQ_clear= 0x184 / sizeof(unsigned),
    ATRQ      = 0x18C / sizeof(unsigned),
    HCC_swres = 1 << 16,
    HCC_lnke  = 1 << 17,
    HCC_lps   = 1 << 19,
  };


void
wait_loop(unsigned *reg, unsigned mask, unsigned value)
{
  while (((*reg) & mask) != value)
    {
      out_char('.');
      wait(1);
    }
}


int
ohci_initialize(unsigned *dev)
{
  // soft reset chip
  dev[HCC_set] = HCC_swres;
  wait_loop(dev + HCC_set, HCC_swres, 0);

  // disable all including byte swapping
  dev[HCC_clear] = ~0U;
  // enable LinkPowerStatus
  dev[HCC_set] = HCC_lps;
  // wait some time to let device enable the link
  wait(50);

  // reset Link Control register
  dev[LC_clear] = 0xffffffff;
  // accept requests from all nodes
  dev[ASRFH_set] = 0x80000000;
  // accept physical requests from all nodes in our local bus
  dev[PHRFH_set] = 0x7fffffff;
  dev[PHRFL_set] = 0xffffffff;
  // allow access up to 0xffff00000000;
  dev[PHUP] = 0xffff0000;
  // we retry because of a busy partner
  dev[ARTRY] = 0x822;
   // enable link
  dev[HCC_set] = HCC_lnke;
  // display link speed and max size of packets
  out_description("BOPT:", dev[BOPT]);
  out_description("HCC:",  dev[HCC_set]);
  return 0;
}


int
send_message(unsigned *dev, short channel, void *data, unsigned datalen)
{

  static unsigned d[12] __attribute__((aligned(16)));

  CHECK3(-10, datalen > (512 - 2*4), "too much to send");

  // output_more_imm descriptor
  d[0] = 0x02000000 | 8;
  // send async stream on channel with 400Mbit
  d[4] = 0xA0 | (channel << 8) |  (2 << 16);
  d[5] = datalen << 16;

  // output_last descriptor
  d[8] = 0x100c0000 | datalen;
  d[9] = (unsigned) data;
  d[10] = 0;
  d[11] = 0;

  // clear run bit
  dev[ATRQ_clear]= 1 << 15;
  wait_loop(dev + ATRQ_clear, 1 << 15, 0);

  // set pointer and run the commands
  dev[ATRQ] = ((sizeof(d)+15)/16) | (unsigned) &d;  
  dev[ATRQ_set] = (1 << 15);
  
  // wait until it is transmitted
  wait_loop(dev + ATRQ_set, 1<<10, 0);
  return 0;
}



int
__main()
{
  out_info(VERSION " init ohci1394 for debugging");
  unsigned device = pci_find_device_per_class(0x0c00);
  CHECK3(10, !device, "no ohci1394 found");

  unsigned *dev = (unsigned *) pci_read_long(device+0x10);
  CHECK3(11, !dev, "ohci1394 registers invalid");

  CHECK3(12, ((*dev) & 0xff00ff) == 0x10001, "invalid version");
  out_description("device", device);
  CHECK3(13, ohci_initialize(dev), "initialize failed");
  while (1)
    {
      send_message(dev, 3, "hello world\n", 13);
      wait(1000);      
    }
  return 0;
}
