/*
 * \brief   fulda_tool - use raw1394 to access the 1394 card
 * \date    2007-04-20
 * \author  Bernhard Kauer <kauer@tudos.org>
 */
/*
 * Copyright (C) 2007  Bernhard Kauer <kauer@tudos.org>
 * Technische Universitaet Dresden, Operating Systems Research Group
 *
 * This file is part of the FULDA package, which is distributed under
 * the  terms  of the  GNU General Public Licence 2.  Please see the
 * COPYING file for details.
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <libraw1394/raw1394.h>


#define CHECK(ERROR, COND, MSG) \
{                       \
  if (COND)             \
  {                     \
    if (errno)          \
       perror(MSG);     \
    else                \
       fprintf(stderr, "%s\n", MSG); \
    exit(ERROR);   \
  }                \
}


enum
  {
    BUS_ID = 0x3ff,
    CSR_BASE = 0xfffff0000000ULL,
  };


static
raw1394handle_t
init_raw1394(int port_nr)
{
  raw1394handle_t res;

  CHECK(-10, !((res = raw1394_new_handle())), "init of raw1394 lib failed");
  int port_count = raw1394_get_port_info(res, NULL, 0);
  CHECK(-11, port_count <= port_nr, "port not available");
  CHECK(-12, raw1394_set_port(res, port_nr), "port config changed");

  return res;
}


static
void
list_nodes(raw1394handle_t handle)
{
  int count = raw1394_get_nodecount(handle);
  int local = raw1394_get_local_id(handle) & 0x3f;

  while (count--)
    {
      quadlet_t uuid[2] = {0, 0};
      int rc;
      
      if (!(rc = raw1394_read(handle, count | (BUS_ID << 6), CSR_BASE + 0x40c, 4, uuid)))
	rc = raw1394_read(handle, count | (BUS_ID << 6), CSR_BASE + 0x410, 4, uuid+1);
      fprintf(stderr, "%#04x: %08x%08x [%s]\n", count | (BUS_ID << 6),  uuid[0], uuid[1], rc ? strerror(errno) : (count == local ? "local" : "remote"));
    }
}


int
my_req_callback(raw1394handle_t handle,
		struct raw1394_arm_request_response *arm_req_resp,
		unsigned int requested_length,
		void *pcontext,
		byte_t request_type)
{
  switch(request_type)
    {
    case RAW1394_ARM_WRITE:
      printf("%s", arm_req_resp->request->buffer);
      break;
    default:
      assert(!"unimplemented"); 
    }
}


enum raw1394_iso_disposition
iso_handler(raw1394handle_t handle,
	    unsigned char *data,
	    unsigned int len,
	    unsigned char channel,
	    unsigned char tag,
	    unsigned char sy,
	    unsigned int cycle,
	    unsigned int dropped)
{
  printf("%s", data);
  fflush(stdout);
  return RAW1394_ISO_OK;
}


int
main(int argc, char **argv)
{
  raw1394handle_t handle;
  char c;

  int mode = -1;
  int port_nr = 0;
  int node = 0;
  int block_size = 4;
  long long addr = 0;
  unsigned long count = 1;


  while ((c = getopt (argc, argv, "hlrwdip:n:a:s:c:")) != -1)
    switch (c)
      {
      case 'h':
      default:
	fprintf(stderr, "Usage: %s [-hlrwdi][n node][-p port][-a addr][-s size][-c count]\n", argv[0]);
	return 1;
      case 'l':
	mode = 0;
	break;
      case 'r':
	mode = 1;
	break;
      case 'w':
	mode = 2;
	break;
      case 'd':
	mode = 3;
	break;
      case 'i':
	mode = 4;
	break;
      case 'p':
	port_nr = atoi(optarg);
	break;
      case 'n':
        node = strtol(optarg, NULL, 0);
	break;
      case 'a':
        addr = strtoll(optarg, NULL, 0);
	break;
      case 's':
        block_size = strtol(optarg, NULL, 0);
	break;
      case 'c':
        count = strtoul(optarg, NULL, 0);
	break;
      }

  handle = init_raw1394(port_nr);
  switch (mode)
    {
    case 0:
      list_nodes(handle);
      break;
    case 1:
    case 2:
      {
	fprintf(stderr, "%s() node %x addr %llx bs: %x count: %lx\n", mode == 1 ? "read" : "write", node, addr, block_size, count);
	CHECK(20, (node < 0) || (node >= raw1394_get_nodecount(handle)), "invalid node number");
	CHECK(21, block_size <= 0, "invalid block size");
	void *buffer = malloc(block_size);
	CHECK(22, !buffer, "could not allocate buffer");
	
	if (mode == 1)
	  {
	    
	    unsigned block;
	    for (block = 0; block < count; block++)
	      {
		CHECK(23, raw1394_read(handle, node | (BUS_ID << 6), addr + block * block_size, block_size, buffer), "read failed");
		write(1, buffer, block_size);
	      }
	  }
	else
	  {
	    unsigned len;
	    if (0 >= ((len = read(0, buffer, block_size))))
	      break;
	    CHECK(24, raw1394_write(handle, node | (BUS_ID << 6), addr, len, buffer), "write failed");
	    addr += len;
	  }
      }
      break;
    case 3:
      
      {
	struct raw1394_arm_reqhandle arm_tag = { &my_req_callback, 0};
	fprintf(stderr, "dump() addr %llx size: %x\n", addr, block_size);
	CHECK(30, raw1394_arm_register (handle, addr, block_size, NULL, 
					(int) &arm_tag,
					0,  
					0,
					RAW1394_ARM_WRITE | RAW1394_ARM_READ),"register range mapping failed");
	while (1)
	    raw1394_loop_iterate(handle);
      }
      break;
    case 4:
      fprintf(stderr, "dump_iso() channel %llx maxsize: %x\n", addr, block_size);
      CHECK(40, (addr < 0) || (addr > 63), "invalid channel number");
      CHECK(41, raw1394_iso_recv_init (handle, 
				       &iso_handler,
				       1,
				       block_size,
				       addr,
				       RAW1394_DMA_PACKET_PER_BUFFER,
				       1), 
	    "init iso recv failed");
      CHECK(42, raw1394_iso_recv_start (handle, -1, -1, 0), "recv start failed");
      while (1)
	    raw1394_loop_iterate(handle);
      break;
    default:
      CHECK(30, 1, "need on of the -{hlrdi} commands");
    }
  
  raw1394_destroy_handle(handle);
  return 0;
}
